
// Copyright Ben Moir 2016
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "vector.h"

static void vector_reserve(vector_t *this, size_t n);
static void vector_insert(vector_t *this, size_t n, void *obj);
static void vector_remove(vector_t *this, size_t n);
static void vector_push_back(vector_t *this, void *obj);
static void vector_pop_back(vector_t *this);
static void *vector_back(vector_t *this);
static void *vector_front(vector_t *this);
static void *vector_at(vector_t *this, size_t at);

vector_t *new_vector(size_t tsize, size_t default_capacity)
{
    vector_t *vec  = malloc(sizeof(vector_t));
    
    vec->push_back = vector_push_back;
    vec->pop_back  = vector_pop_back;
    vec->back      = vector_back;
    vec->front     = vector_front;
	vec->reserve   = vector_reserve;
	vec->insert    = vector_insert;
	vec->remove    = vector_remove;
	vec->at        = vector_at;
    
    vec->capacity  = default_capacity >= 2 ? default_capacity : 10;
    vec->tsize     = tsize;
    vec->count     = 0;
    vec->data      = (void *)calloc(vec->capacity, vec->tsize);
    return vec;
}

static void vector_reserve(vector_t *this, size_t n)
{
	if (n > this->capacity)
	{
		this->capacity = n;
		this->data = realloc(this->data, this->tsize * this->capacity);
	}
}

static void vector_insert(vector_t *this, size_t n, void *obj)
{
	if (n >= this->count)
		return;
	if (n >= this->capacity)
		this->reserve(this, n + 1);
	
	// split data into two left hand and right hand sides of n
	// insert new object in between left hand and right hand sides
	memmove((unsigned char *)this->data, (unsigned char *)this->data, this->tsize * n);
	memcpy((unsigned char *)this->data + (this->tsize * n), obj, this->tsize);
	memmove((unsigned char *)this->data + (this->tsize * n + 1), (unsigned char *)this->data + (this->tsize * (n + 1)), this->tsize * (this->count - n - 1));
	
	this->count++;
}

static void vector_remove(vector_t *this, size_t n)
{
	if (this->count == 0)
		return;
	if (n >= this->count)
		return;
	
	// split data into two left hand and right hand sides of n
	memmove((unsigned char *)this->data, (unsigned char *)this->data, this->tsize * n);
	memmove((unsigned char *)this->data + (this->tsize * n), (unsigned char *)this->data + (this->tsize * (n + 1)), this->tsize * (this->count - n - 1));
	
	this->count--;
}

static void vector_push_back(vector_t *this, void *obj)
{
    if (++this->count > this->capacity)
		this->reserve(this, this->count);
    memcpy((unsigned char *)this->data + (this->tsize * (this->count - 1)), obj, this->tsize);
}

static void vector_pop_back(vector_t *this)
{
    if (this->count > 0)
	{
        this->count--;
	}
}

static void *vector_back(vector_t *this)
{
    if (this == NULL || this->count == 0)
        return NULL;
    else
        return (unsigned char *)this->data + (this->tsize * (this->count - 1));
}

static void *vector_front(vector_t *this)
{
    if (this == NULL || this->count == 0)
        return NULL;
    else
        return this->data;
}

static void *vector_at(vector_t *this, size_t at)
{
	if (at >= this->count || this->count == 0)
		return NULL;
	return (unsigned char *)this->data + (this->tsize * at);
}
